/*
 * MultiMail offline mail reader
 * Interface, ShadowedWin, ListWindow

 Copyright (c) 1996 Kolossvary Tamas <thomas@tvnet.hu>
 Copyright (c) 1997 John Zero <john@graphisoft.hu>
 Copyright (c) 1998 William McBrine <wmcbrine@clark.net>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */

#include "interfac.h"
#include "isoconv.h"
#include "mysystem.h"

Win::Win(int y1, int x1, int y2, int x2, chtype backg)
{
	win = newwin(y1, x1, y2, x2);
	Clear(backg);

	keypad(win, TRUE);
	nodelay(win, FALSE);
	cursor_off();
}

Win::~Win()
{
	delwin(win);
}

void Win::Clear(chtype backg)
{
	wbkgdset(win, backg | ' ');
	werase(win);
	wbkgdset(win, ' ');
	wattrset(win, backg);
}

void Win::putch(int x, int y, chtype z)
{
	mvwaddch(win, x, y, z);
}

void Win::putchnstr(int x, int y, const chtype *z, int len)
{
	mvwaddchnstr(win, x, y, (chtype *) z, len);
}

void Win::putstring(int x, int y, const char *z)
{
	mvwaddstr(win, x, y, (char *) z);
}

void Win::attrib(chtype x)
{
	wattrset(win, x);
}

void Win::horizline(int y, int len)
{
	wmove(win, y, 1);
	whline(win, ACS_HLINE, len);
}

void Win::update()
{
	wrefresh(win);
}

void Win::delay_update()
{
	wnoutrefresh(win);
}

void Win::wtouch()
{
	touchwin(win);
	wnoutrefresh(win);
}

void Win::wscroll(int i)
{
	scrollok(win, TRUE);
	wscrl(win, i);
	scrollok(win, FALSE);
}

void Win::cursor_on()
{
	leaveok(win, FALSE);
}

void Win::cursor_off()
{
	leaveok(win, TRUE);
}

int Win::keypressed()
{
	nodelay(win, TRUE);
	return (wgetch(win) != ERR);
}

int Win::inkey()
{
	int c;

	do
		c = wgetch(win);
	while (c == ERR);
	return c;
}

void Win::boxtitle(chtype backg, const char *title, chtype titleAttrib)
{
	wattrset(win, backg);
	box(win, 0, 0);

	if (title) {
		putch(0, 2, ACS_RTEE);
		putch(0, 3 + strlen(title), ACS_LTEE);
		wattrset(win, titleAttrib);
		putstring(0, 3, title);
		wattrset(win, backg);
	}
}

ShadowedWin::ShadowedWin(int y1, int x1, int y2, int x2, chtype backg,
	const char *title, chtype titleAttrib) : Win(y1, x1, y2, x2, backg)
{
#ifdef USE_SHADOWS
	int i, sx, sy;
	chtype *right, *lower;
# ifndef NCURSES_VERSION
#  ifdef __PDCURSES__
	WINDOW *newscr = curscr;
#  else
	WINDOW *newscr = stdscr;
#  endif
# endif
	int limit = 2 + (interface->active() != letter);
	sx = x1;
	sy = y1;
	while ((sx + x2) > (COLS - limit))
		sx--;
	while ((sy + y2) > (LINES - limit))
		sy--;

	right = new chtype[(sy - 1) << 1];
	lower = new chtype[sx + 1];
	for (i = 0; i < (sy - 1); i++) {
		right[i << 1] = (mvwinch(newscr, (y2 + i + 1),
			(x2 + sx)) & (A_CHARTEXT | A_ALTCHARSET));
		right[(i << 1) + 1] = (mvwinch(newscr, (y2 + i + 1),
			(x2 + sx + 1)) & (A_CHARTEXT | A_ALTCHARSET));
	}
	mvwinchnstr(newscr, (y2 + sy), (x2 + 2), lower, sx);
	shadow = newwin(sy, sx, y2 + 1, x2 + 2);
	leaveok(shadow, TRUE);
	for (i = 0; i < (sy - 1); i++) {
		mvwaddch(shadow, i, sx - 2, (right[i << 1] |
			C_SHADOW | A_BOLD));
		mvwaddch(shadow, i, sx - 1, (right[(i << 1) + 1] |
			C_SHADOW | A_BOLD));
	}
	for (i = 0; i < sx; i++)
		mvwaddch(shadow, sy - 1, i, (lower[i] & (A_CHARTEXT |
			A_ALTCHARSET)) | C_SHADOW | A_BOLD);
	wnoutrefresh(shadow);
#endif
	boxtitle(backg, title, titleAttrib);

#ifdef USE_SHADOWS
	delete[] lower;
	delete[] right;
#endif
}

ShadowedWin::~ShadowedWin()
{
#ifdef USE_SHADOWS
	delwin(shadow);
#endif
}

void ShadowedWin::touch()
{
#ifdef USE_SHADOWS
	touchwin(shadow);
	wnoutrefresh(shadow);
#endif
	Win::wtouch();
}

int ShadowedWin::getstring(int y, int x, char *string, int n, int bg_color, 
		int fg_color)
{
	int i, j, end, c;
	char *tmp = new char[n];

	cropesp(string);

	wattrset(win, fg_color);
	for (i = 0; i < n; i++) {
		putch(y, x + i, ACS_BOARD);
		tmp[i] = '\0';
	}
	putstring(y, x, string);

	cursor_on();
	//keypad(win, TRUE);
	//nodelay(win, FALSE);

	update();

	i = end = 0;
	while (!end) {
		c = inkey();
		switch (c) {
		case KEY_DOWN:
			end++;
		case KEY_UP:
			end++;
		case '\r':
		case '\n':
			end++;
		case '\033':
			end++;
			break;
		case KEY_LEFT:
			if (!tmp[0]) {
				strcpy(tmp, string);
				i = strlen(string);
			}
			if (i > 0)
				i--;
			break;
		case KEY_RIGHT:
			if ((i < n - 1) && tmp[i])
				i++;
			break;
		case 127:
		case KEY_DC:	// Delete key 
			strncpy(&tmp[i], &tmp[i + 1], n - i - 1);
			tmp[n - 1] = '\0';
			break;
		case 8:	// Ctrl-H
		case KEY_BACKSPACE:
			if (i > 0) {
				strncpy(&tmp[i - 1], &tmp[i], n - i - 1);
				tmp[n - 1] = '\0';
				i--;
			}
			break;
		default:
			for (j = n - 1; j > i; j--)
				tmp[j] = tmp[j - 1];
			tmp[i] = c;
			if (i < n - 1)
				i++;
		}
		for (j = 0; j < n; j++)
			putch(y, x + j, (tmp[j] ? tmp[j] : ACS_BOARD));
		wmove(win, y, x + i);
		update();
	}

	if (tmp[0])
		strcpy(string, tmp);

	wattrset(win, bg_color);

	j = strlen(string);
	for (i = 0; i < n; i++)
		putch(y, x + i, ((i < j) ? string[i] : ' '));

	update();
	cursor_off();

	delete[] tmp;

	return end - 1;
}

InfoWin::InfoWin(int y1, int x1, int y2, int x2, chtype backg,
	const char *title, chtype titleAttrib, int bottx, int topx) :
	ShadowedWin(y1, x1, y2, x2, backg, title, titleAttrib)
{
	lineBuf = new char[COLS];
	info = new Win(y1 - bottx, x1 - 2, y2 + topx, x2 + 1, backg);
}

InfoWin::~InfoWin()
{
	delete info;
	delete[] lineBuf;
}

void InfoWin::irefresh()
{
	info->delay_update();
}

void InfoWin::touch()
{
	ShadowedWin::touch();
	info->wtouch();
}

void InfoWin::oneline(int i, chtype ch)
{
	charconv_in(lineBuf);
	info->attrib(ch);
	info->putstring(i, 0, lineBuf);
}

void InfoWin::iscrl(int i)
{
	info->wscroll(i);
}

ListWindow::ListWindow()
{
	position = active = 0;
	oldPos = oldActive = -100;
}

void ListWindow::checkPos(int limit)
{
	if (active >= limit)
		active = limit - 1;

	if (active < 0)
		active = 0;

	if (active < position)
		position = active;
	else
		if (active - position >= list_max_y)
			position = active - list_max_y + 1;

	if (limit > list_max_y) {
		if (position < 0)
			position = 0;
		else
			if (position > limit - list_max_y)
				position = limit - list_max_y;
	} else
		position = 0;
}

void ListWindow::Draw()
{
	int i, j, limit = NumOfItems();

	checkPos(limit);

	i = position - oldPos;
	switch (i) {
	case -1:
	case 1:
		list->iscrl(i);
	case 0:
		if (active != oldActive) {
			j = oldActive - position;
			if ((j >= 0) && (j < list_max_y))
				oneLine(j);
		}
		oneLine(active - position);
		list->irefresh();
		break;
	default:
		for (j = 0; j < list_max_y; j++) {
			oneLine(j);
			if (position + j == limit - 1)
				j = list_max_y;
		}
		ReDraw();
	}
	oldPos = position;
	oldActive = active;
}

void ListWindow::Move(direction dir)
{
	int limit = NumOfItems();

	switch (dir) {
	case UP:
		active--;
		break;
	case DOWN:
		active++;
		break;
	case PGUP:
		position -= list_max_y;
		active -= list_max_y;
		break;
	case PGDN:
		position += list_max_y;
		active += list_max_y;
		break;
	case HOME:
		active = 0;
		break;
	case END:
		active = limit - 1;
		break;
	}
	checkPos(limit);
}

void ListWindow::ReDraw()
{
	list->touch();
}

void ListWindow::KeyHandle(int key)
{
	int draw = 1;

	switch (key) {
	case KEY_DOWN:
		Move(DOWN);
		break;
	case KEY_UP:
		Move(UP);
		break;
	case KEY_HOME:
		Move(HOME);
		break;
	case KEY_END:
		Move(END);
		break;
	case 'B':
	case KEY_PPAGE:
		Move(PGUP);
		break;
	case ' ':
	case 'F':
	case KEY_NPAGE:
		Move(PGDN);
		break;
	default:
		draw = 0;
		extrakeys(key);
	}
	if (draw)
		Draw();
}

Interface::Interface()
{
        isoconvert =
#if defined (__MSDOS__) || defined (__EMX__)
        0;
#else
        1;
#endif
}

Interface::~Interface()
{
	delete screen;
	touchwin(stdscr);
	refresh();
	leaveok(stdscr, FALSE);
	echo();
	endwin();
}

void Interface::init_colors()
{
	for (int y = COLOR_BLACK; y <= COLOR_WHITE; y++)
		for (int x = COLOR_BLACK; x <= COLOR_WHITE; x++)
			init_pair(((x << 3) + y), x, y);

	// Color pair 0 cannot be set (argh!), so swap:

	init_pair(((COLOR_WHITE << 3) + COLOR_WHITE),
		COLOR_BLACK, COLOR_BLACK);

	// Of course, now I can't do white-on-white (grey). :-(
}

void Interface::screen_init()
{
	initscr();
	refresh();
	start_color();
	init_colors();
	cbreak();
	noecho();
	nonl();

	// Make a new background window, fill it with ACS_BOARD characters:
	screen = new Win(0, 0, 0, 0, C_SBACK | ACS_BOARD);

	if ((COLS < 60) || (LINES < 20))
		fatalError("A screen at least 60x20 is required");

	// Border and title:
	char tmp[80];
	sprintf(tmp, MM_TOPHEADER, MM_NAME, MM_MAJOR, MM_MINOR);
	screen->boxtitle(C_SBORDER | A_BOLD, tmp, C_SBACK | A_BOLD);

	// Help window area:
	screen->attrib(C_SSEPBOTT);
	screen->horizline(LINES - 5, COLS - 2);
	screen->delay_update();
}

int Interface::WarningWindow(const char *warning, const char *yes,
				const char *no)
{
	int x, c, result = 0, def_val = 1;	//possible values: 1 or 0

	x = strlen(warning) + 4;

	ShadowedWin warn(7, x, LINES / 2 - 4, (COLS - x) / 2, C_WTEXT
		| A_BOLD);

	warn.putstring(2, 2, warning);
	warn.putstring(4, x / 3 + 1, &yes[1]);
	warn.putstring(4, 2 * x / 3 - strlen(no) + 1, &no[1]);
	warn.attrib(C_WTEXTHI | A_BOLD);
	warn.putch(4, x / 3, yes[0]);
	warn.putch(4, 2 * x / 3 - strlen(no), no[0]);

	while (!result) {
		warn.putstring(4, x / 3 - 1, def_val ? "[" : " ");
		warn.putstring(4, x / 3 + strlen(yes), def_val ? "]" : " ");
		warn.putstring(4, 2 * x / 3 - strlen(no) - 1, def_val ?
			" " : "[");
		warn.putstring(4, 2 * x / 3, def_val ? " " : "]");
		warn.update();

		c = warn.inkey();
		if ((c == tolower(yes[0])) || (c == toupper(yes[0]))) {
			def_val = 1;
			result = 1;
		} else if ((c == tolower(no[0])) || (c == toupper(no[0]))) {
			def_val = 0;
			result = 1;
		} else
			switch (c) {
			case KEY_LEFT:
			case KEY_RIGHT:
			case 9:
				def_val = !def_val;
				break;
			case '\r':
			case '\n':
				result = 1;
			}
	}
	return def_val;
}

statetype Interface::active()
{
	return state;
}

void Interface::oldstate(statetype n)
{
	helpwindow.Delete();

	switch (n) {
	case packetlist:
		packets.Delete();
		break;
	case arealist:
		areas.Delete();
		break;
	case letterlist:
		letters.Delete();
		break;
	case address:
		addresses.Delete();
		oldstate(prevstate);
		break;
	case tagwin:
		taglines.Delete();
		oldstate(prevstate);
		break;
	case littlearealist:
		delete littleareas;
	case letter_help:
	case letter:
		letterwindow.Delete();
		break;
	case ansiwin:
		ansiwindow.Delete();
	default:;
	}
}

void Interface::newstate(statetype n)
{
	statetype oldst = state;
	state = n;

	switch (n) {
	case packetlist:
		screen->wtouch();
		currList = &packets;
		packets.MakeActive();
		break;
	case arealist:
		screen->wtouch();
		currList = &areas;
		areas.MakeActive();
		break;
	case letterlist:
		screen->wtouch();
		currList = &letters;
		letters.MakeActive();
		break;
	case letter:
		letterwindow.MakeActive(oldst == letterlist);
		break;
	case letter_help:
		letterwindow.MakeActive(0);
		break;
	case littlearealist:
		letterwindow.MakeActive(0);
		littleareas = new LittleAreaListWindow(Key);
		currList = littleareas;
		break;
	case address:
		newstate(prevstate);
		state = address;
		currList = &addresses;
		addresses.MakeActive(1);	// how to determine param?
		break;
	case tagwin:
		newstate(prevstate);
		Tagwin();
		break;
	case ansiwin:
		ansiwindow.MakeActive();
	default:;
	}

	helpwindow.MakeActive();
}

void Interface::changestate(statetype n)
{
	oldstate(state);
	newstate(n);
}

int Interface::select()
{
	switch (state) {
	case packetlist:
		unsaved_reply = any_read = 0;
		packets.OpenPacket();
		changestate(arealist);
		areas.FirstUnread();
		break;
	case arealist:
		areas.Select();
		if (mm.areaList->getNoOfLetters() > 0) {
			mm.areaList->getLetterList();
			changestate(letterlist);
			letters.FirstUnread();
		}
		break;
	case letterlist:
		letters.Select();
		changestate(letter);
		any_read = 1;
		break;
	case letter:
		letterwindow.Next();
		break;
	case letter_help:
		back();
		break;
	case littlearealist:
		changestate(letter);
		if (mm.areaList->isNetmail())
			if ((Key == 'N') || (Key == 'E'))
				addressbook();
			else {
				net_address nm = letterwindow.PickNetAddr();
				letterwindow.set_Letter_Params(&nm, NULL);
			}
		Key = '\0';
		letterwindow.EnterLetter();
		changestate(letter);
		break;
	case address:
	case tagwin:
		currList->KeyHandle('\n');
		return back();
	default:;
	}
	return 0;
}

int Interface::back()
{
	switch (state) {
	case packetlist:
		if (WarningWindow("Do you really want to quit?",
				  "Yes", "No")) {
			oldstate(state);
			return 1;
		} else
			changestate(packetlist);
		break;
	case arealist:
		if (any_read) {
			if (WarningWindow("Save lastread pointers?",
					  "Yes", "No"))
				mm.saveRead();
			any_read = 0;
		}
		if (unsaved_reply)
			if (WarningWindow(
		"The content of the reply area has changed. Create new packet?",
						 "Yes", "No"))
				create_reply_packet();
		mm.Delete();
		changestate(packetlist);
		break;
	case letterlist:
		delete mm.letterList;
		changestate(arealist);
		break;
	case littlearealist:
		delete littleareas;
		state = letter;
		areas.Select();
	case letter_help:
		changestate(letter);
		break;
	case letter:
		changestate(letterlist);
		break;
	case address:
	case tagwin:
		changestate(prevstate);
		return 1;
	case ansiwin:
		changestate(prevstate);
	default:;
	}
	return 0;
}

#ifdef SIGWINCH
void Interface::sigwinch()
{
	oldstate(state);

	delete screen;
	endwin();
	screen_init();

	newstate(state);
	doupdate();
}
#endif

void Interface::kill_letter()
{
	if (WarningWindow(
		"Are you sure you want to delete this letter?",
			  "Yes", "No")) {

		mm.areaList->killLetter(mm.letterList->getMsgNum());
		unsaved_reply = 1;

		changestate(letterlist);
		if (!mm.areaList->getNoOfLetters())
			back();
	} else
		changestate(state);
}

void Interface::create_reply_packet()
{
	mm.areaList->makeReply();
	unsaved_reply = 0;
}

void Interface::addressbook(int NoEnter)
{
	prevstate = state;
	addresses.MakeActive(NoEnter);
	state = address;
	currList = &addresses;
	KeyHandle();
}

void Interface::Tagwin()
{
	prevstate = state;
	taglines.MakeActive();
	state = tagwin;
	currList = &taglines;
}

void Interface::ansiLoop(const char *source, const char *title)
{
	prevstate = state;
	ansiwindow.set(source, title);
	changestate(ansiwin);
}

void Interface::TransToggle()
{
	isoconvert = !isoconvert;
	changestate(state);
}

void Interface::setUnsaved()
{
	unsaved_reply = 1;
}

void Interface::setAnyRead()
{
	any_read = 1;
}

void Interface::KeyHandle()		// Main loop
{
	int c, end = 0;

	while (!end) {
		doupdate();

		c = screen->inkey();

		if ((c >= 'a') && (c <= 'z'))
			c = toupper(c);

		switch (c) {
		case 'Q':
		case '\033':	//the escape key
		case KEY_BACKSPACE:
			end = back();
			break;
		case '\n':
		case '\r':
			end = select();
			break;
		default:
			switch (state) {
			case letter:
				Key = c;
				letterwindow.KeyHandle(c);
				break;
			case letter_help:
				back();
				break;
			case ansiwin:
				ansiwindow.KeyHandle(c);
				break;
			default:
				currList->KeyHandle(c);
			}
		}
	}
}
